/* ------------------------------------------------------------------------------------------------------------------------------
    vs10xx device functions.
    Copyright (C) 2010 Richard van Paasen <rvpaasen@t3i.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   ----------------------------------------------------------------------------------------------------------------------------- */

#include "vs1053_common.h"
#include "vs1053_iocomm.h"
#include "vs1053_queue.h"
#include "vs1053_device.h"
#include "vs1053_plugins.h"

#include <linux/delay.h>
#include <linux/kthread.h>

/* set clockf value */
static int clockf = 0xe000;
module_param(clockf, int, 0644);

/* load new plugin */
static int plugin = 1;
module_param(plugin, int, 0644);

struct vs1053_device_t {
	int id;
	int open;
	int valid;
	int start;
	int finish;
	struct mutex lock;
	struct device *dev;
	struct task_struct *kthread;
	wait_queue_head_t wq;
	unsigned long underrun;
	int version;
};

static struct vs1053_device_t vs1053_device[VS1053_MAX_DEVICES];

/* ----------------------------------------------------------------------------------------------------------------------------- */
/* VS1053 DEVICE READ/WRITE SCI REGISTERS                                                                                        */
/* ----------------------------------------------------------------------------------------------------------------------------- */

static int vs1053_device_w_sci_reg(int id, unsigned char reg, unsigned char msb, unsigned char lsb) {

	int status;

	unsigned char cmd[] = {0x02, reg, msb, lsb};

	status = vs1053_io_ctrl_xf(id, cmd, sizeof(cmd), NULL, 0);

	vs1053_nsy("id:%d %02X %02X%02X", id, (int)cmd[1], (int)cmd[2], (int)cmd[3]);

	if (!vs1053_io_wtready(id, 10)) {

		vs1053_err("id:%d timeout", id);
		status = -1;
	}

	return status;
}

static int vs1053_device_r_sci_reg(int id, unsigned char scir, unsigned char* msb, unsigned char* lsb) {

	int status;

	unsigned char cmd[] = {0x03, scir};
	unsigned char res[] = {0x00, 0x00};

	status = vs1053_io_ctrl_xf(id, cmd, sizeof(cmd), res, sizeof(res));

	*msb = res[0];
	*lsb = res[1];

	vs1053_nsy("id:%d %02X %02X%02X", id, (int)cmd[1], (int)res[0], (int)res[1]);

	if (!vs1053_io_wtready(id, 10)) {

		vs1053_err("id:%d timeout", id);
		status = -1;
	}

	return status;
}


/* ----------------------------------------------------------------------------------------------------------------------------- */
/* VS1053 DEVICE LOAD PLUGIN                                                                                                     */
/* ----------------------------------------------------------------------------------------------------------------------------- */

#define vs1053_plugin vs1053_plugin_140_noflac

static int vs1053_device_loadplugin(int id) {

	int i = 0, n = sizeof(vs1053_plugin)/sizeof(vs1053_plugin[0]), status = 0;

	vs1053_dbg("start:%d", id);

	/* load a compressed plugin */
	while (status == 0 && i < n) {
		unsigned short addr, m, val;
		addr = vs1053_plugin[i++];
		m = vs1053_plugin[i++];
		if (m & 0x8000U) {
			/* RLE run */
			m &= 0x7FFF;
			val = vs1053_plugin[i++];
			while (status == 0 && m--) {
				/* replicate m samples */
				status = vs1053_device_w_sci_reg(id, addr, (val >> 8) & 0x00ff, val & 0x00ff);
			}
		} else {
			/* Copy run */
			while (status == 0 && m--) {
				 /* copy m samples */
				val = vs1053_plugin[i++];
				status = vs1053_device_w_sci_reg(id, addr, (val >> 8) & 0x00ff, val & 0x00ff);
			}
		}
	}

	vs1053_inf("id:%d loaded plugin (%d/%d bytes)", id, i, n);

	if (i < n) {

		vs1053_err("id:%d plugin not loaded completely", id);
	}

	vs1053_dbg("done:%d", id);

	return status;
}

/* ----------------------------------------------------------------------------------------------------------------------------- */
/* VS1053 DEVICE RESET                                                                                                           */
/* ----------------------------------------------------------------------------------------------------------------------------- */

static void vs1053_device_hwreset(int id) {

	vs1053_dbg("start:%d", id);

	vs1053_io_reset(id);

	if (!vs1053_io_wtready(id, 50)) {

		vs1053_wrn("id:%d timeout", id);
	}

	vs1053_dbg("done:%d", id);
}

static int vs1053_device_swreset(int id, int testmode) {

	int status = 0;

	vs1053_dbg("start:%d", id);

	if (testmode) {

		// WRITE MODE SM_SDINEW | SM_RESET | SM_TESTS
		status = vs1053_device_w_sci_reg(id, 0x00, 0x08, 0x24);

	} else {

		// WRITE MODE SM_SDINEW | SM_RESET
		status = vs1053_device_w_sci_reg(id, 0x00, 0x08, 0x04);
		//status = vs1053_device_w_sci_reg(id, 0x00, 0x08, 0x00);
	}

	vs1053_dbg("done:%d", id);

	return status;
}

/* ----------------------------------------------------------------------------------------------------------------------------- */
/* VS1053 DEVICE OPERATIONS                                                                                                      */
/* ----------------------------------------------------------------------------------------------------------------------------- */

int vs1053_device_reset(int id) {

	int status;

	vs1053_dbg("start:%d", id);

	mutex_lock(&vs1053_device[id].lock);

	vs1053_queue_flush(id);

	vs1053_device_hwreset(id);

	status = vs1053_device_swreset(id, 0);

	if (status != 0) {

		vs1053_device_hwreset(id);
		status = vs1053_device_swreset(id, 0);
	}

	if (status == 0 && clockf != 0) {

		status = vs1053_device_w_sci_reg(id, 0x03, (clockf & 0xff00) >> 8, (clockf & 0xff));
	}

	if (status == 0) {

		unsigned char msb, lsb;
		status = vs1053_device_r_sci_reg(id, 0x01, &msb, &lsb);
		vs1053_device[id].version = (lsb & 0xf0) >> 4;
	}

	if (status == 0 && plugin == 1) {

		status = vs1053_device_loadplugin(id);
	}

	if (status == 0) {

		unsigned char buffer[2] = { 0, 0 };
		status = vs1053_io_data_tx(id, buffer, 2);
	}

	mutex_unlock(&vs1053_device[id].lock);

	vs1053_dbg("done:%d", id);

	return status;
}

int vs1053_device_sinetest(int id) {

	int status;

	vs1053_inf("id:%d started sine test", id);

	mutex_lock(&vs1053_device[id].lock);

	vs1053_queue_flush(id);

	vs1053_device_hwreset(id);

	status = vs1053_device_swreset(id, 1);

	if (status != 0) {

		vs1053_device_hwreset(id);
		status = vs1053_device_swreset(id, 1);
	}

	if (status == 0) {

		// GENERATE SINE
		unsigned char test[] = {0x53, 0xef, 0x6e, 0xa3, 0x00, 0x00, 0x00, 0x00};
		status = vs1053_io_data_tx(id, test, sizeof(test));
	}

	mutex_unlock(&vs1053_device[id].lock);

	return status;
}


int vs1053_device_memtest(int id) {

	int status;
	unsigned char msb, lsb;

	vs1053_inf("id:%d started mem test", id);

	mutex_lock(&vs1053_device[id].lock);

	vs1053_queue_flush(id);

	vs1053_device_hwreset(id);

	status = vs1053_device_swreset(id, 1);

	if (status != 0) {

		vs1053_device_hwreset(id);
		status = vs1053_device_swreset(id, 0);
	}


	if (status == 0) {

		// TEST MEMORY
		unsigned char test[] = {0x4d, 0xea, 0x6d, 0x54, 0x00, 0x00, 0x00, 0x00};
		status = vs1053_io_data_tx(id, test, sizeof(test));
	}

	if (status == 0) {

		msleep(VS1053_MSEC(1100000));

		// READ HDAT0
		status = vs1053_device_r_sci_reg(id, 0x08, &msb, &lsb);
		msb = msb & 0x83; /* bits 14:10 are unsued */
	}

	mutex_unlock(&vs1053_device[id].lock);

	if (status == 0) {

		vs1053_inf("id:%d mem test complete - result:%02X%02X", id, (int)msb, (int)lsb);
	}

	return status;
}


int vs1053_device_getscireg(int id, struct vs1053_scireg *scireg) {

	int status;

	mutex_lock(&vs1053_device[id].lock);

	status = vs1053_device_r_sci_reg(id, scireg->reg & 0x0F, &scireg->msb, &scireg->lsb);

	mutex_unlock(&vs1053_device[id].lock);

	return status;
}


int vs1053_device_setscireg(int id, struct vs1053_scireg *scireg) {

	int status;

	mutex_lock(&vs1053_device[id].lock);

	status = vs1053_device_w_sci_reg(id, scireg->reg & 0x0F, scireg->msb, scireg->lsb);

	mutex_unlock(&vs1053_device[id].lock);

	return status;
}

/*
int vs1053_device_getclockf(int id, struct vs1053_clockf *clkf) {

	int status = 0;
	unsigned char msb, lsb;

	mutex_lock(&vs1053_device[id].lock);

	status = vs1053_device_r_sci_reg(id, 0x03, &msb, &lsb);

	mutex_unlock(&vs1053_device[id].lock);

	clkf->mul = (msb >> 5) & 0x07;
	clkf->add = (msb >> 3) & 0x03;
	clkf->clk = (msb & 0x07) << 8 | lsb;

	return status;
}


int vs1053_device_setclockf(int id, struct vs1053_clockf *clkf) {

	int status = 0;
	unsigned char msb = ((clkf->mul & 0x07) << 5) | ((clkf->add & 0x03) << 3) | ((clkf->clk >> 8) & 0x07);
	unsigned char lsb = clkf->clk & 0xff;

	mutex_lock(&vs1053_device[id].lock);

	status = vs1053_device_w_sci_reg(id, 0x03, msb, lsb);

	mutex_unlock(&vs1053_device[id].lock);

	return status;
}


int vs1053_device_getvolume(int id, struct vs1053_volume *volume) {

	int status = 0;
	unsigned char msb, lsb;

	mutex_lock(&vs1053_device[id].lock);

	status = vs1053_device_r_sci_reg(id, 0x0B, &msb, &lsb);

	mutex_unlock(&vs1053_device[id].lock);

	volume->left = 255 - msb;
	volume->rght = 255 - lsb;

	return status;
}


int vs1053_device_setvolume(int id, struct vs1053_volume *volume) {

	int status = 0;
	unsigned char msb = 255 - (volume->left & 0xff);
	unsigned char lsb = 255 - (volume->rght & 0xff);

	mutex_lock(&vs1053_device[id].lock);

	status = vs1053_device_w_sci_reg(id, 0x0B, msb, lsb);

	mutex_unlock(&vs1053_device[id].lock);

	return status;
}


int vs1053_device_gettone(int id, struct vs1053_tone *tone) {

	int status = 0;
	unsigned char msb, lsb;

	mutex_lock(&vs1053_device[id].lock);

	status = vs1053_device_r_sci_reg(id, 0x02, &msb, &lsb);

	mutex_unlock(&vs1053_device[id].lock);

	tone->trebboost = (msb >> 4) & 0x0f;
	tone->treblimit = msb & 0x0f;
	tone->bassboost = (lsb >> 4) & 0x0f;
	tone->basslimit = lsb & 0x0f;

	return status;
}


int vs1053_device_settone(int id, struct vs1053_tone *tone) {

	int status = 0;
	unsigned char msb = ((tone->trebboost & 0x0f) << 4) | (tone->treblimit & 0x0f);
	unsigned char lsb = ((tone->bassboost & 0x0f) << 4) | (tone->basslimit & 0x0f);

	mutex_lock(&vs1053_device[id].lock);

	status = vs1053_device_w_sci_reg(id, 0x02, msb, lsb);

	mutex_unlock(&vs1053_device[id].lock);

	return status;
}
*/
/* ----------------------------------------------------------------------------------------------------------------------------- */
/* VS1053 DEVICE THREAD                                                                                                          */
/* ----------------------------------------------------------------------------------------------------------------------------- */

static inline void vs1053_device_setopen(int id) {

	mutex_lock(&vs1053_device[id].lock);
	vs1053_device[id].open = 1;
	mutex_unlock(&vs1053_device[id].lock);
}

static inline void vs1053_device_clropen(int id) {

	mutex_lock(&vs1053_device[id].lock);
	vs1053_device[id].open = 0;
	mutex_unlock(&vs1053_device[id].lock);

	wake_up(&vs1053_device[id].wq);
}

static inline int vs1053_device_getopen(int id) {

	return vs1053_device[id].open;
}

static inline void vs1053_device_setpause(int id) {

	mutex_lock(&vs1053_device[id].lock);
	vs1053_device[id].start = 0;
	mutex_unlock(&vs1053_device[id].lock);
}

static inline void vs1053_device_clrpause(int id) {

	mutex_lock(&vs1053_device[id].lock);
	vs1053_device[id].start = 1;
	mutex_unlock(&vs1053_device[id].lock);

	wake_up(&vs1053_device[id].wq);
}

static inline int vs1053_device_getpause(int id) {

	return (vs1053_device[id].start ? 0 : 1);
}

static inline void vs1053_device_setfinish(int id) {

	mutex_lock(&vs1053_device[id].lock);
	vs1053_device[id].finish = 1;
	mutex_unlock(&vs1053_device[id].lock);
}

static inline void vs1053_device_clrfinish(int id) {

	mutex_lock(&vs1053_device[id].lock);
	vs1053_device[id].finish = 0;
	mutex_unlock(&vs1053_device[id].lock);

	wake_up(&vs1053_device[id].wq);
}

static inline int vs1053_device_getfinish(int id) {

	return vs1053_device[id].finish;
}


static int vs1053_device_kthread(void *arg) {

	struct vs1053_device_t *device = (struct vs1053_device_t*)arg;

	do {

		//int status = 0;

		if (vs1053_device_getpause(device->id)) {

			/* on hold, wait for start */
			wait_event_timeout(device->wq, device->start, msecs_to_jiffies(10));

		}
		else if (vs1053_queue_isempty(device->id)) {

			/* no data --> pause */
			if (vs1053_device_getopen(device->id)) {

				vs1053_nsy("id:%d no data", device->id);
				vs1053_device_setpause(device->id);

				if (!vs1053_device_getfinish(device->id)) {

					/* log underrun */
					device->underrun++;
				}
			}

		} else if (!vs1053_io_isready(device->id)) {

			/* io not ready to receive */
			vs1053_nsy("id:%d not ready", device->id);
			vs1053_io_wtready(device->id, 10);

		} else {

			mutex_lock(&device->lock);

			while (vs1053_io_isready(device->id) && !vs1053_queue_isempty(device->id)) {

				/* transmit data */
				struct vs1053_queue_buf_t *buffer;
				buffer = vs1053_queue_gethead(device->id);
				/* status = */vs1053_io_data_tx(device->id, buffer->data, buffer->len);
				vs1053_queue_dequeue(device->id);
			}

			mutex_unlock(&device->lock);
		}

	} while (!kthread_should_stop());

	return 0;
}

/* ----------------------------------------------------------------------------------------------------------------------------- */
/* VS1053 DEV INTERFACE                                                                                                          */
/* ----------------------------------------------------------------------------------------------------------------------------- */

int vs1053_device_isvalid(int id) {

	return !(id < 0 || id >= VS1053_MAX_DEVICES || !vs1053_device[id].valid);
}

int vs1053_device_open(int id) {

	if (vs1053_device_getopen(id)) {
		return -1;
	}

	vs1053_device_setopen(id);

	return 0;
}

int vs1053_device_release(int id) {

	struct vs1053_scireg scireg_addr = { 0x07, 0x1e, 0x06 };
	struct vs1053_scireg scireg_data = { 0x06, 0x00, 0x00 };
	//int status;

	vs1053_device_clrpause(id);

	while (!vs1053_queue_isempty(id)) {

		msleep(1);
	}

	/* status = */vs1053_device_setscireg(id, &scireg_addr);
	/* status = */vs1053_device_getscireg(id, &scireg_data);

	{
		struct vs1053_queue_buf_t buffer;
		unsigned char msb, lsb;
		int i;

		buffer.len = sizeof(buffer.data);
		memset(buffer.data, scireg_data.lsb, buffer.len);

		for (i = 0; i < 65; i++) {

			/* status = */vs1053_io_data_tx(id, buffer.data, buffer.len);
		}

		/*status = */vs1053_device_w_sci_reg(id, 0x00, 0x08, 0x08);

		for (i = 0; i < 64; i++) {

			/*status = */vs1053_io_data_tx(id, buffer.data, buffer.len);

			vs1053_io_wtready(id, 10);

			/*status = */vs1053_device_r_sci_reg(id, 0x00, &msb, &lsb);

			if ((lsb & 0x08) == 0) {

				break;
			}
		}

		if ((lsb & 0x08) == 1) {

			vs1053_device_swreset(id, 0);
		}
	}

	vs1053_device_clropen(id);

	return 0;
}


struct vs1053_queue_buf_t* vs1053_device_getbuf(int id) {

	struct vs1053_queue_buf_t *buffer = NULL;

	if (!vs1053_queue_isfull(id)) {
		mutex_lock(&vs1053_device[id].lock);
		buffer = vs1053_queue_getslot(id);
		mutex_unlock(&vs1053_device[id].lock);
	}

	return buffer;
}

int vs1053_device_write(int id) {

	//int status = 0;

	mutex_lock(&vs1053_device[id].lock);
	vs1053_queue_enqueue(id);
	mutex_unlock(&vs1053_device[id].lock);

	if (vs1053_device_getpause(id) && vs1053_queue_isfull(id)) {

		vs1053_device_clrpause(id);
	}

	return 0;//status;
}

int vs1053_device_getfree(int id) {

	return vs1053_queue_getfree(id);
}

int vs1053_device_status(int id, char* buf) {

	char *status = vs1053_device[id].start ? (vs1053_device[id].finish ? "finishing" : "playing") : "stopped";

	return sprintf(buf, "version: %d\nunderrun: %lu\nstatus: %s\n",
		vs1053_device[id].version,
		vs1053_device[id].underrun,
		status
	);
}

/* ----------------------------------------------------------------------------------------------------------------------------- */
/* VS1053 DEVICE INIT/EXIT                                                                                                       */
/* ----------------------------------------------------------------------------------------------------------------------------- */

int __init vs1053_device_init(int id, struct device *dev) {

	int status;

	vs1053_dbg("start:%d", id);

	vs1053_device[id].id = id;
	vs1053_device[id].open = 0;
	vs1053_device[id].valid = 0;
	vs1053_device[id].start = 0;
	vs1053_device[id].finish = 0;
	vs1053_device[id].underrun = 0;
	vs1053_device[id].version = -1;

	vs1053_device[id].dev = dev;

	/* initialize mutex */
	mutex_init(&vs1053_device[id].lock);

	/* initialize wait queue */
	init_waitqueue_head(&vs1053_device[id].wq);

	/* reset device */
	status = vs1053_device_reset(id);

	if (status == 0) {

		/* start device thread */
		vs1053_device[id].kthread = kthread_run(vs1053_device_kthread, &vs1053_device[id], "%s-%d", VS1053_NAME, id);

		if (vs1053_device[id].kthread == NULL) {

			vs1053_err("id:%d kthread_run", id);
			status = -1;
		}
	}

	if (status == 0) {

		/* flag valid */
		vs1053_device[id].valid = 1;
	}

	vs1053_dbg("done:%d", id);

	return status;
}

void __exit vs1053_device_exit(int id) {

	if (vs1053_device[id].kthread != NULL) {

		/* stop device thread */
		kthread_stop(vs1053_device[id].kthread);
	}
}
