/* ------------------------------------------------------------------------------------------------------------------------------
    vs10xx io functions.
    Copyright (C) 2010 Richard van Paasen <rvpaasen@t3i.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   ----------------------------------------------------------------------------------------------------------------------------- */

#include "vs1053_common.h"
#include "vs1053_iocomm.h"

#include <linux/gpio.h>
#include <linux/delay.h>
#include <linux/spi/spi.h>
#include <linux/interrupt.h>
#include <linux/sched.h>

#ifdef AS_MODULE
#include "board.h"
#endif

/* interrupt mode */
static int irqmode = 1;
module_param(irqmode, int, 0644);

/* signore dreq status */
static int skipdreq = 0;
module_param(skipdreq, int, 0644);

struct vs1053_chip {
	int dreq_val;
	int gpio_reset;
	int gpio_dreq;
	int irq_dreq;
	struct spi_device *spi_ctrl;
	struct spi_device *spi_data;
	wait_queue_head_t wq;
};

static struct vs1053_chip vs1053_chips[VS1053_MAX_DEVICES];

/* ----------------------------------------------------------------------------------------------------------------------------- */
/* VS1053 SPI MISC                                                                                                               */
/* ----------------------------------------------------------------------------------------------------------------------------- */

static irqreturn_t vs1053_io_irq(int irq, void *arg) {
/*
 *  Descr:  interrupt handler, record DREQ value and wakeup waiting task
 *  Return: irq handled
 */

	int id = (int)arg;

	vs1053_chips[id].dreq_val = gpio_get_value(vs1053_chips[id].gpio_dreq);

	if (vs1053_chips[id].dreq_val) {
		wake_up(&vs1053_chips[id].wq);
	}

	return IRQ_HANDLED;
}

int vs1053_io_isready(int id) {
/*
 *  Descr:  Test if DREQ is high
 *  Return: 1 --> DREQ is high (ready)
 *          0 --> DREQ is low (busy)
 */

	int dreq_val = (irqmode ? vs1053_chips[id].dreq_val : gpio_get_value(vs1053_chips[id].gpio_dreq));

	vs1053_nsy("id:%d dreq_val:%d", id, dreq_val);

	return (skipdreq ? 1 : dreq_val);
}

int vs1053_io_wtready(int id, unsigned timeout) {
/*
 *  Descr:  Wait timeout [msec] for DREQ to become high
 *  Return: 1 --> DREQ is high (ready)
 *          0 --> timeout (busy)
 */

	if (irqmode) {

		if (!vs1053_chips[id].dreq_val) {

			wait_event_timeout(vs1053_chips[id].wq, vs1053_chips[id].dreq_val, msecs_to_jiffies(timeout));
		}

	} else {

		unsigned long end = jiffies + msecs_to_jiffies(timeout);

		while ( !vs1053_io_isready(id) && (jiffies < end) ) {

			msleep(1);
		}
	}

	return vs1053_io_isready(id);
}

void vs1053_io_reset(int id) {
/*
 *  Descr:  Reset by pulling down gpio
 *  Return: -
 */
	if(vs1053_chips[id].gpio_reset >= 0)
	{
		gpio_set_value(vs1053_chips[id].gpio_reset, 0);
		udelay(5);
		gpio_set_value(vs1053_chips[id].gpio_reset, 1);
		udelay(1000);
	}
}

/* ----------------------------------------------------------------------------------------------------------------------------- */
/* VS1053 SPI TRANSFER                                                                                                           */
/* ----------------------------------------------------------------------------------------------------------------------------- */

int vs1053_io_ctrl_xf(int id, const char *txbuf, unsigned txlen, char *rxbuf, unsigned rxlen) {

	struct spi_device *device = vs1053_chips[id].spi_ctrl;
	struct spi_message	spi_mesg;
	struct spi_transfer	spi_xfer_tx;
	struct spi_transfer	spi_xfer_rx;
	int status;

	spi_message_init(&spi_mesg);

	if (txbuf) {
		memset(&spi_xfer_tx, 0, sizeof spi_xfer_tx);
		spi_xfer_tx.tx_buf = txbuf;
		spi_xfer_tx.len = txlen;
		spi_xfer_tx.delay_usecs = rxbuf ? 0 : 10;
		spi_message_add_tail(&spi_xfer_tx, &spi_mesg);
	}

	if (rxbuf) {
		memset(&spi_xfer_rx, 0, sizeof spi_xfer_rx);
		spi_xfer_rx.rx_buf = rxbuf;
		spi_xfer_rx.len = rxlen;
		spi_xfer_rx.delay_usecs = 10;
		spi_message_add_tail(&spi_xfer_rx, &spi_mesg);
	}

	status = spi_sync(device, &spi_mesg);
	if (status < 0) {
		vs1053_err("id:%d spi_sync failed", id);
	}

	return status;
}

int vs1053_io_data_rx(int id, char *rxbuf, unsigned rxlen) {

	struct spi_device *device = vs1053_chips[id].spi_data;
	struct spi_message	spi_mesg;
	struct spi_transfer	spi_xfer_rx;
	int status;

	spi_message_init(&spi_mesg);

	memset(&spi_xfer_rx, 0, sizeof spi_xfer_rx);
	spi_xfer_rx.rx_buf = rxbuf;
	spi_xfer_rx.len = rxlen;
	//spi_xfer_rx.delay_usecs = 0;
	spi_message_add_tail(&spi_xfer_rx, &spi_mesg);

	status = spi_sync(device, &spi_mesg);
	if (status < 0) {
		vs1053_err("id:%d spi_sync failed", id);
	}

	return status;
}

int vs1053_io_data_tx(int id, const char *txbuf, unsigned txlen) {

	struct spi_device *device = vs1053_chips[id].spi_data;
	struct spi_message	spi_mesg;
	struct spi_transfer	spi_xfer_tx;
	int status;

	spi_message_init(&spi_mesg);

	memset(&spi_xfer_tx, 0, sizeof spi_xfer_tx);
	spi_xfer_tx.tx_buf = txbuf;
	spi_xfer_tx.len = txlen;
	//spi_xfer_tx.delay_usecs = 0;
	spi_message_add_tail(&spi_xfer_tx, &spi_mesg);

	status = spi_sync(device, &spi_mesg);
	if (status < 0) {
		vs1053_err("id:%d spi_sync failed", id);
	}

	return status;
}

/* ----------------------------------------------------------------------------------------------------------------------------- */
/* VS1053 SPI PROBES                                                                                                             */
/* ----------------------------------------------------------------------------------------------------------------------------- */

static int vs1053_spi_ctrl_probe(struct spi_device *spi) {

	int status = 0;
#ifndef AS_MODULE
	struct vs1053_board_info* boardinfo = (struct vs1053_board_info*) spi->dev.platform_data;

	if (boardinfo == NULL) {
		vs1053_err("no board info provided for vs1053-ctrl device");
		status = -1;
	}

	if (status == 0) {
		if (boardinfo->device_id < 0 || boardinfo->device_id >= VS1053_MAX_DEVICES) {
			vs1053_err("id out of range:%d", boardinfo->device_id);
			status = -1;
		}
	}

	if (status == 0) {
		vs1053_chips[boardinfo->device_id].spi_ctrl = spi;
		vs1053_chips[boardinfo->device_id].gpio_reset = boardinfo->gpio_reset;
		vs1053_chips[boardinfo->device_id].gpio_dreq = boardinfo->gpio_dreq;

		vs1053_inf("id:%d ctrl spi:%d.%d gpio_reset:%d gpio_dreq:%d",
			boardinfo->device_id, spi->master->bus_num, spi->chip_select, boardinfo->gpio_reset, boardinfo->gpio_dreq);
	}
#endif
	return status;
}


static struct spi_driver vs1053_spi_ctrl = {
	.driver = {
		.name		= VS1053_SPI_CTRL,
		.owner		= THIS_MODULE,
	},
	.probe		= vs1053_spi_ctrl_probe,
//	.remove		= __devexit_p(vs1053_ctrl_remove),
};


static int vs1053_spi_data_probe(struct spi_device *spi) {

	int status = 0;
#ifndef AS_MODULE
	struct vs1053_board_info* boardinfo = (struct vs1053_board_info*) spi->dev.platform_data;

	if (boardinfo == NULL) {
		vs1053_err("no board info provided for vs1053-data device");
		status = -1;
	}

	if (status == 0) {
		if (boardinfo->device_id < 0 || boardinfo->device_id >= VS1053_MAX_DEVICES) {
			vs1053_err("id out of range:%d", boardinfo->device_id);
			status = -1;
		}
	}

	if (status == 0) {
		vs1053_chips[boardinfo->device_id].spi_data = spi;

		vs1053_inf("id:%d data spi:%d.%d gpio_reset:%d gpio_dreq:%d",
			boardinfo->device_id, spi->master->bus_num, spi->chip_select, boardinfo->gpio_reset, boardinfo->gpio_dreq);
	}
#endif
	return status;
}


static struct spi_driver vs1053_spi_data = {
	.driver = {
		.name		= VS1053_SPI_DATA,
		.owner		= THIS_MODULE,
	},
	.probe		= vs1053_spi_data_probe,
//	.remove		= __devexit_p(vs1053_data_remove),
};


/* ----------------------------------------------------------------------------------------------------------------------------- */
/* VS1053 IO REG/UNREG                                                                                                           */
/* ----------------------------------------------------------------------------------------------------------------------------- */
#ifdef AS_MODULE
static int __init add_device_to_bus(int i)
{
	struct spi_device* spi_device;
	struct spi_board_info* spi_info = &spi_vs1053_devices[i];
	struct vs1053_board_info* boardinfo =
		(struct vs1053_board_info*) spi_info->platform_data;
	struct spi_master* spi_master =
		spi_busnum_to_master(spi_info->bus_num);

	if (!spi_master) {
		vs1053_err("spi_busnum_to_master(%d) returned NULL\n",
								spi_info->bus_num);
		return -1;
	}
	if (!strcmp(spi_info->modalias, VS1053_SPI_CTRL)) {
		spi_device = spi_new_device(spi_master, spi_info);
		if (!spi_device) {
			vs1053_err("spi_new_device\n");
			return -1;
		}
		vs1053_chips[boardinfo->device_id].spi_ctrl = spi_device;
		vs1053_chips[boardinfo->device_id].gpio_reset = boardinfo->gpio_reset;
		vs1053_chips[boardinfo->device_id].gpio_dreq = boardinfo->gpio_dreq;

		vs1053_inf("id:%d ctrl spi:%d.%d gpio_reset:%d gpio_dreq:%d",
				boardinfo->device_id, spi_device->master->bus_num, spi_device->chip_select, boardinfo->gpio_reset, boardinfo->gpio_dreq);
	} else if (!strcmp(spi_info->modalias, VS1053_SPI_DATA)) {
		spi_device = spi_new_device(spi_master, spi_info);
		if (!spi_device) {
			vs1053_err("spi_new_device\n");
			return -1;
		}
		vs1053_chips[boardinfo->device_id].spi_data = spi_device;

		vs1053_inf("id:%d data spi:%d.%d gpio_reset:%d gpio_dreq:%d",
			boardinfo->device_id, spi_device->master->bus_num, spi_device->chip_select, boardinfo->gpio_reset, boardinfo->gpio_dreq);
	} /* else other device */
	return 0;
}
#endif
int __init vs1053_io_register(void) {

	int s1, s2;
	int i;

	/* register vs1053 ctrl spi driver */
	s1 = spi_register_driver(&vs1053_spi_ctrl);
	if (s1 < 0) {
		vs1053_err("spi_register_driver: ctrl");
	}

	/* register vs1053 data spi driver */
	s2 = spi_register_driver(&vs1053_spi_data);
	if (s2 < 0) {
		vs1053_err("spi_register_driver: data");
	}
#ifdef AS_MODULE
	for (i = 0; i < ARRAY_SIZE(spi_vs1053_devices); i++) {
		add_device_to_bus(i);
	}
#endif
	return (s1==0 && s2==0 ? 0  : -1);
}


void __exit vs1053_io_unregister(void) {

	/* unregister spi devices */
	spi_unregister_driver(&vs1053_spi_ctrl);
	spi_unregister_driver(&vs1053_spi_data);
}

/* ----------------------------------------------------------------------------------------------------------------------------- */
/* VS1053 IO INIT/EXIT                                                                                                           */
/* ----------------------------------------------------------------------------------------------------------------------------- */

int __init vs1053_io_init(int id) {

	struct vs1053_chip *chip;
	int status = -1;

	if (id < 0 || id >= VS1053_MAX_DEVICES) {

		vs1053_err("id:%d out of range", id);

	} else {

		chip = &vs1053_chips[id];

		if (chip->spi_ctrl == NULL || chip->spi_data == NULL) {

			vs1053_dbg("id:%d no board config", id);

		} else {

			/* initialize wait queue */
			init_waitqueue_head(&chip->wq);
			/* request reset signal */
			if(chip->gpio_reset < 0){
				status = 0;
			} else {
				status = gpio_request(chip->gpio_reset, "vs1053_reset");
				if (status < 0) {
					vs1053_err("gpio_request gpio_reset:%d", chip->gpio_reset);
					chip->gpio_reset = -1;
				}

				if (status == 0) {
					/* set gpio_reset as output, keep vs1053 in reset */
					status = gpio_direction_output(chip->gpio_reset, 0);
					if (status < 0) {
						vs1053_err("gpio_direction_output gpio_reset:%d", chip->gpio_reset);
					} else {
						gpio_set_value(chip->gpio_reset, 0);
					}
				}
			}
			if (status == 0) {
				/* request dreq signal */
				status = gpio_request(chip->gpio_dreq, "vs1053_dreq");
				if (status < 0) {
					vs1053_err("gpio_request gpio_dreq:%d", chip->gpio_dreq);
					chip->gpio_dreq = -1;
				}
			}

			if (status == 0) {
				/* set gpio_dreq as input */
				status = gpio_direction_input(chip->gpio_dreq);
				if (status < 0) {
					vs1053_err("gpio_direction_input gpio_dreq:%d", chip->gpio_dreq);
				} else {
					chip->dreq_val = gpio_get_value(chip->gpio_dreq);
				}
			}

			if (irqmode) {
				if (status == 0) {
					/* request dreq irq */
					chip->irq_dreq = gpio_to_irq(chip->gpio_dreq);
					if (chip->irq_dreq < 0) {
						vs1053_err("gpio_to_irq gpio_dreq:%d", chip->gpio_dreq);
						status = -1;
					}
				}

				if (status == 0) {
					/* request irq for gpio_dreq */
					status = request_irq(chip->irq_dreq, vs1053_io_irq, IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING, VS1053_NAME, (void*)id);
					if (status < 0) {
						vs1053_err("request_irq irq_dreq:%d", chip->irq_dreq);
						chip->irq_dreq = -1;
					}
				}
			}
		}
	}

	if (status == 0) {
		vs1053_inf("id:%d gpio_reset:%d gpio_dreq:%d irq_dreq:%d", id, chip->gpio_reset, chip->gpio_dreq, chip->irq_dreq);
	}

	return status;
}

void __exit vs1053_io_exit(int id) {

	struct vs1053_chip *chip = &vs1053_chips[id];

	/* release dreq irq */
	if (irqmode && chip->irq_dreq > 0) {
		synchronize_irq(chip->irq_dreq);
		free_irq(chip->irq_dreq, (void*)id);
	}

	/* free gpio dreq */
	if (chip->gpio_dreq > 0) {
	    gpio_free(chip->gpio_dreq);
	}

	/* free gpio reset */
	if (chip->gpio_reset > 0) {
		gpio_set_value(chip->gpio_reset, 0);
	    gpio_free(chip->gpio_reset);
	}
#ifdef AS_MODULE
	/* free devices */
	if (chip->spi_ctrl) {
		spi_unregister_device(chip->spi_ctrl);
	}
	if (chip->spi_data) {
		spi_unregister_device(chip->spi_data);
	}
#endif
}

