/* ------------------------------------------------------------------------------------------------------------------------------
    vs10xx main functions.
    Copyright (C) 2010 Richard van Paasen <rvpaasen@t3i.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   ----------------------------------------------------------------------------------------------------------------------------- */

#include "vs1053_common.h"
#include "vs1053_device.h"
#include "vs1053_queue.h"
#include "vs1053_iocomm.h"

#include "adav801_main.h"
#include "cue_switch_main.h"

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include <linux/cdev.h>
#include <linux/kthread.h>
#include <linux/delay.h>

static int debug = 0;
int *vs1053_debug = &debug;
module_param(debug, int, 0644);

MODULE_DESCRIPTION("device driver for vs1053");
MODULE_AUTHOR("Richard van Paasen");
MODULE_LICENSE("GPL v2");

static dev_t vs1053_cdev_region;
static struct cdev *vs1053_cdev;

/* ----------------------------------------------------------------------------------------------------------------------------- */
/* CHARDEV INTERFACE                                                                                                             */
/* ----------------------------------------------------------------------------------------------------------------------------- */

static int vs1053_open(struct inode *inode, struct file *file) {

	int id = MINOR(inode->i_rdev);
	int status;

	vs1053_dbg("id:%d", id);

	status = vs1053_device_open(id);

	if (status < 0) {
		vs1053_inf("id:%d not valid or already open", id);
		status = -EACCES;
	} else {
		file->private_data = (void*)id;
	}

	return status;
}


static int vs1053_release(struct inode *inode, struct file *file) {

	int id = (int)file->private_data;
	int status;

	vs1053_dbg("id:%d", id);

	status = vs1053_device_release(id);

	if (status < 0) {
		status = -EIO;
	}

	return status;
}


static ssize_t vs1053_write(struct file *file, const char __user *usrbuf, size_t lbuf, loff_t *ppos) {

	int id = (int)file->private_data;
	int status = 0;

	int nbytes;
	int acttodo;
	int copied = 0;

	struct vs1053_queue_buf_t *buffer;

	vs1053_nsy("id:%d", id);

	do {

		buffer = vs1053_device_getbuf(id);

		if ((buffer == NULL)) {

			vs1053_dbg("id:%d queue full", id);
			msleep(1);

		} else {

			acttodo = sizeof(buffer->data) > (lbuf-copied) ? (lbuf-copied) : sizeof(buffer->data);
			nbytes = acttodo - copy_from_user(buffer->data, usrbuf+copied, acttodo);
			buffer->len = nbytes;
			status = vs1053_device_write(id);
			copied += nbytes;
		}

	} while (copied < lbuf && buffer != NULL);

	vs1053_nsy("id:%d copied %d bytes", id, copied);

	return (status < 0 ? status : copied);
}


static long vs1053_ioctl(struct file *file, unsigned int cmd, unsigned long arg) {

	int type =_IOC_TYPE(cmd), nr = _IOC_NR(cmd), /*dir = _IOC_DIR(cmd),*/ size = _IOC_SIZE(cmd);
	int id = (int)file->private_data;
	char __user * usrbuf = (void __user *)arg;

	struct vs1053_scireg scireg;
	struct vs1053_clockf clockf;
	struct vs1053_volume volume;
	struct vs1053_tone tone;

	if (type != VS1053_CTL_TYPE) {
        //vs1053_wrn("id:%d invalid ioctl type:%c nr:%d", id, type, nr);
        return -ENOTTY;//-EINVAL;
	}

	switch (nr) {
		case _IOC_NR(VS1053_CTL_RESET):
			vs1053_device_reset(id);
			break;
		case _IOC_NR(VS1053_CTL_SINETEST):
			vs1053_device_sinetest(id);
			break;
		case _IOC_NR(VS1053_CTL_MEMTEST):
			vs1053_device_memtest(id);
			break;
		case _IOC_NR(VS1053_CTL_GETSCIREG):
			if (copy_from_user(&scireg, usrbuf, size)) return -EFAULT;
			vs1053_device_getscireg(id, &scireg);
			if (copy_to_user(usrbuf, &scireg, size)) return -EFAULT;
			break;
		case _IOC_NR(VS1053_CTL_SETSCIREG):
			if (copy_from_user(&scireg, usrbuf, size)) return -EFAULT;
			vs1053_device_setscireg(id, &scireg);
			break;
/*		case _IOC_NR(VS1053_CTL_GETCLOCKF):
			vs1053_device_getclockf(id, &clockf);
			if (copy_to_user(usrbuf, &clockf, size)) return -EFAULT;
			break;
		case _IOC_NR(VS1053_CTL_SETCLOCKF):
			if (copy_from_user(&clockf, usrbuf, size)) return -EFAULT;
			vs1053_device_setclockf(id, &clockf);
			break;
		case _IOC_NR(VS1053_CTL_GETVOLUME):
			vs1053_device_getvolume(id, &volume);
			if (copy_to_user(usrbuf, &volume, size)) return -EFAULT;
			break;
		case _IOC_NR(VS1053_CTL_SETVOLUME):
			if (copy_from_user(&volume, usrbuf, size)) return -EFAULT;
			vs1053_device_setvolume(id, &volume);
			break;
		case _IOC_NR(VS1053_CTL_GETTONE):
			vs1053_device_gettone(id, &tone);
			if (copy_to_user(usrbuf, &tone, size)) return -EFAULT;
			break;
		case _IOC_NR(VS1053_CTL_SETTONE):
			if (copy_from_user(&tone, usrbuf, size)) return -EFAULT;
			vs1053_device_settone(id, &tone);
			break;
*/		default:
	        vs1053_wrn("id:%d invalid ioctl type:%c nr:%d", id, type, nr);
	        return -EINVAL;
	}

	return 0;
}

static const struct file_operations vs1053_fops = {
	.owner = THIS_MODULE,
	.open = vs1053_open,
	.release = vs1053_release,
	.write = vs1053_write,
	.unlocked_ioctl = vs1053_ioctl,
};

/* ----------------------------------------------------------------------------------------------------------------------------- */
/* VS1053 SYSFS INTERFACE                                                                                                        */
/* ----------------------------------------------------------------------------------------------------------------------------- */

static struct class_attribute vs1053_class_attrs[] = {
	__ATTR_NULL,
};

static struct class vs1053_class = {
	.name =		VS1053_NAME,
	.owner =	THIS_MODULE,
	.class_attrs =	vs1053_class_attrs,
};

static ssize_t vs1053_sys_reset(struct device *dev, struct device_attribute *attr, const char *buf, size_t size) {

	const int id = (int)dev_get_drvdata(dev);
	vs1053_device_reset(id);
	return size;
}

static ssize_t vs1053_sys_test(struct device *dev, struct device_attribute *attr, const char *buf, size_t size) {

	const int id = (int)dev_get_drvdata(dev);
	vs1053_device_sinetest(id);
	return size;
}

static ssize_t vs1053_sys_status(struct device *dev, struct device_attribute *attr, char *buf) {

	const int id = (int)dev_get_drvdata(dev);
	return vs1053_device_status(id, buf);
}

static const DEVICE_ATTR(reset, 0222, NULL, vs1053_sys_reset);
static const DEVICE_ATTR(test, 0222, NULL, vs1053_sys_test);
static const DEVICE_ATTR(status, 0444, vs1053_sys_status, NULL);

static const struct attribute *vs1053_attrs[] = {
	&dev_attr_reset.attr,
	&dev_attr_test.attr,
	&dev_attr_status.attr,
	NULL,
};

static const struct attribute_group vs1053_attr_group = {
	.attrs = (struct attribute **) vs1053_attrs,
};


/* ----------------------------------------------------------------------------------------------------------------------------- */
/* MODULE INIT/EXIT                                                                                                              */
/* ----------------------------------------------------------------------------------------------------------------------------- */

static int __init vs1053_init (void) {

	int status = 0;
	struct device *dev = NULL;
	int i;

	vs1053_dbg("start");

	snd_master_clock_init();
	switch_init();

	if (status == 0) {
		/* register vs1053 region */
		status = alloc_chrdev_region(&vs1053_cdev_region, 0, VS1053_MAX_DEVICES, VS1053_NAME);
		if (status < 0) {
			vs1053_err("alloc_chrdev_region");
		}
	}

	if (status == 0) {
		/* register vs1053 class */
		status = class_register(&vs1053_class);
		if (status < 0) {
			vs1053_err("class_register");
		}
	}

	if (status == 0) {
		/* create vs1053 character device driver */
		vs1053_cdev = cdev_alloc();
		if (vs1053_cdev == NULL) {
			vs1053_err("cdev_alloc");
			status = -1;
		} else {
			cdev_init(vs1053_cdev, &vs1053_fops);
			status = cdev_add(vs1053_cdev, vs1053_cdev_region, VS1053_MAX_DEVICES);
			if (status < 0) {
				vs1053_err("cdev_add");
			}
		}
	}

	if (status == 0) {
		/* register io */
		status = vs1053_io_register();
	}

	/* initialize vs1053 devices */
	for (i = 0; (status == 0) && (i < VS1053_MAX_DEVICES); i++) {

		/* init io path */
		if (vs1053_io_init(i) == 0) {

			/* init queue */
			status = vs1053_queue_init(i);

			if (status == 0) {
				/* create vs1053 char device */
				dev = device_create(&vs1053_class, NULL, MKDEV(MAJOR(vs1053_cdev_region), i), (void*)i, "%s-%d", VS1053_NAME, i);
				if (dev == NULL) {
					vs1053_err("device_create id:%d name:%s-%d maj:%d min:%d", i, VS1053_NAME, i, MAJOR(vs1053_cdev_region), i);
					status = -1;
				} else {
					dev_set_drvdata(dev, (void*)i);
					status = sysfs_create_group(&dev->kobj, &vs1053_attr_group);
				}
			}
			
			if (status == 0) {
				status = vs1053_device_init(i, dev);
			}

			if (status == 0) {
				vs1053_inf("id:%d name:%s-%d maj:%d min:%d", i, VS1053_NAME, i, MAJOR(vs1053_cdev_region), i);
			}
		}
	}

	vs1053_dbg("done");

	return status;
}

static void __exit vs1053_exit (void) {

	int i;

	vs1053_dbg("start");

	// TODO: cleanup does not check error states properly

	/* cleanup vs1053 devices */
	for (i = 0; i < VS1053_MAX_DEVICES; i++) {

		if (vs1053_device_isvalid(i)) {

			/* exit device */
			vs1053_device_exit(i);

			/* destroy device */
			device_destroy(&vs1053_class, MKDEV(MAJOR(vs1053_cdev_region), i));

			/* exit queue */
			vs1053_queue_exit(i);

			/* exit io */
			vs1053_io_exit(i);
		}
	}

	vs1053_io_unregister();

	/* cleanup additional devices */
	snd_master_clock_exit();
	switch_exit();

	/* cleanup char device driver */
	if (vs1053_cdev) {
		cdev_del(vs1053_cdev);
	}

	/* unregister vs1053 class */
	class_unregister(&vs1053_class);

	/* unregister vs1053 region */
	unregister_chrdev_region(vs1053_cdev_region, VS1053_MAX_DEVICES);

	vs1053_dbg("done");
}

module_init(vs1053_init);
module_exit(vs1053_exit);

