/* ------------------------------------------------------------------------------------------------------------------------------
    vs10xx queue functions.
    Copyright (C) 2010 Richard van Paasen <rvpaasen@t3i.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   ----------------------------------------------------------------------------------------------------------------------------- */

#include "vs1053_common.h"
#include "vs1053_queue.h"

#include <linux/slab.h>

/* queue size*/
static int queuelen = 8192;
module_param(queuelen, int, 0644);

struct vs1053_queue_entry_t {
	struct list_head list;
	struct vs1053_queue_buf_t buf;
};

struct vs1053_queue_t {
	int id;
	int free;
	struct list_head buf_pool;
	struct list_head buf_data;
};

static struct vs1053_queue_t vs1053_queue[VS1053_MAX_DEVICES];

/* ----------------------------------------------------------------------------------------------------------------------------- */
/* VS1053 QUEUE BUFFERS                                                                                                          */
/* ----------------------------------------------------------------------------------------------------------------------------- */

static int __init vs1053_queue_alloc(int id) {

	struct vs1053_queue_t *queue = &vs1053_queue[id];
	struct vs1053_queue_entry_t *entry;
	int i, status = 0;

	/* create buf_pool list */
	INIT_LIST_HEAD(&queue->buf_pool);

	/* create buf_data list */
	INIT_LIST_HEAD(&queue->buf_data);

	/* allocate buffers in pool */
	for(i = 0; (status == 0) && (i < queuelen); i++) {
		entry = kzalloc(sizeof *entry, GFP_KERNEL);
		if (!entry) {
			vs1053_err("kzalloc queue entry");
			status = -1;
		} else {
			list_add(&entry->list, &queue->buf_pool);
		}
	}

	queue->free = ((status == 0) ? queuelen : 0);

	vs1053_inf("id:%d allocated %d/%d buffers (%d bytes)", queue->id, i, queuelen, (i * sizeof(entry->buf)));

	return status;
}

void vs1053_queue_flush(int id) {

	struct vs1053_queue_t *queue = &vs1053_queue[id];
	struct vs1053_queue_entry_t *entry, *n;

	/* move all buffers to pool */
	if (!list_empty(&queue->buf_data)) {
		list_for_each_entry_safe(entry, n, &queue->buf_data, list) {
			vs1053_dbg("id:%d flush queue entry", id);
			list_move_tail(&entry->list, &queue->buf_pool);
		}
	}
}

static void __exit vs1053_queue_free(int id) {

	struct vs1053_queue_t *queue = &vs1053_queue[id];
	struct vs1053_queue_entry_t *entry, *n;

	vs1053_queue_flush(id);

	/* free buffers */
	list_for_each_entry_safe(entry, n, &queue->buf_pool, list) {
		list_del(&entry->list);
		kfree(entry);
	}

	queue->free = 0;
}

int vs1053_queue_isfull(int id) {

	struct vs1053_queue_t *queue = &vs1053_queue[id];

	return list_empty(&queue->buf_pool);

}

int vs1053_queue_isempty(int id) {

	struct vs1053_queue_t *queue = &vs1053_queue[id];

	return list_empty(&queue->buf_data);

}

int vs1053_queue_getfree(int id) {

	struct vs1053_queue_t *queue = &vs1053_queue[id];

	return queue->free;
}

struct vs1053_queue_buf_t* vs1053_queue_getslot(int id) {

	struct vs1053_queue_t *queue = &vs1053_queue[id];
	/* get the first (empty) data entry from the data queue */
	struct vs1053_queue_entry_t *entry = list_first_entry(&queue->buf_pool, struct vs1053_queue_entry_t, list);

	return &entry->buf;
}

struct vs1053_queue_buf_t* vs1053_queue_gethead(int id) {

	struct vs1053_queue_t *queue = &vs1053_queue[id];
	/* get the first (filled) data entry from the data queue */
	struct vs1053_queue_entry_t *entry = list_first_entry(&queue->buf_data, struct vs1053_queue_entry_t, list);

	return &entry->buf;
}

void vs1053_queue_enqueue(int id) {

	struct vs1053_queue_t *queue = &vs1053_queue[id];
	/* could ask the caller for the entry, but let's do not to prevent queue corruption */
	struct vs1053_queue_entry_t *entry = list_first_entry(&queue->buf_pool, struct vs1053_queue_entry_t, list);
	if (entry->buf.len > 0) {
		list_move_tail(&entry->list, &queue->buf_data);
		queue->free -= 1;
	}
}

void vs1053_queue_dequeue(int id) {

	struct vs1053_queue_t *queue = &vs1053_queue[id];
	/* could ask the caller for the entry, but let's do not to prevent queue corruption */
	struct vs1053_queue_entry_t *entry = list_first_entry(&queue->buf_data, struct vs1053_queue_entry_t, list);
	list_move_tail(&entry->list, &queue->buf_pool);
	entry->buf.len = 0;
	queue->free += 1;
}

/* ----------------------------------------------------------------------------------------------------------------------------- */
/* VS1053 QUEUE INIT/EXIT                                                                                                        */
/* ----------------------------------------------------------------------------------------------------------------------------- */

int __init vs1053_queue_init(int id) {

	int status;

	vs1053_queue[id].id = id;
	status = vs1053_queue_alloc(id);

	return status;
}

void __exit vs1053_queue_exit(int id) {

	vs1053_queue_free(id);

}
